//-------------------------------------------------------------------------------------------------------------------------------------------------------------
//
// Copyright 2024 Apple Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
//-------------------------------------------------------------------------------------------------------------------------------------------------------------

#include "TextureLoader.h"
#import <MetalKit/MetalKit.h>

MTL::Texture* newTextureFromFile(const std::string& texturePath, MTL::Device* pDevice)
{
    NSURL* url = [NSURL fileURLWithPath:[NSString stringWithUTF8String:texturePath.c_str()]];
    MTKTextureLoader* loader = [[MTKTextureLoader alloc] initWithDevice:(__bridge id<MTLDevice>)pDevice];
    
    NSError* __autoreleasing error = nil;
    id<MTLTexture> texture = [loader newTextureWithContentsOfURL:url
                                                         options:@{MTKTextureLoaderOptionLoadAsArray : @(YES), /* Shader converter pipelines expect texture arrays */
                                                                   MTKTextureLoaderOptionTextureStorageMode : @(MTLStorageModePrivate),
                                                                   MTKTextureLoaderOptionTextureUsage : @(MTLTextureUsageShaderRead),
                                                                   MTKTextureLoaderOptionGenerateMipmaps : @(YES)}
                                                           error:&error];
    
    if (!texture)
    {
        NSLog(@"Error loading texture at \"%s\": %@", texturePath.c_str(), error.localizedDescription);
    }
    assert(texture);
    
    return (__bridge_retained MTL::Texture *)texture; // transfer ownership to metal-cpp
}

MTL::Texture* newTextureArrayFromFiles(const std::vector<std::string>& texturePaths, MTL::Device* pDevice, MTL::CommandBuffer* pCommandBuffer)
{
    // Load textures:
    std::vector<NS::SharedPtr<MTL::Texture>> textures;
    for (size_t i = 0; i < texturePaths.size(); ++i)
    {
        textures.push_back(NS::TransferPtr(newTextureFromFile(texturePaths[i], pDevice)));
    }
    
    MTL::Texture* pTexture = nullptr;
    
    // Combine textures into a single texture array:
    
    if (textures.size() > 0)
    {
        auto pTextureDesc = NS::TransferPtr(MTL::TextureDescriptor::alloc()->init());
        
        pTextureDesc->setWidth(textures[0]->width());
        pTextureDesc->setHeight(textures[0]->height());
        pTextureDesc->setPixelFormat(textures[0]->pixelFormat());
        
        pTextureDesc->setTextureType(MTL::TextureType2DArray);
        pTextureDesc->setUsage(MTL::TextureUsageShaderRead);
        pTextureDesc->setStorageMode(MTL::StorageModePrivate);
        pTextureDesc->setArrayLength(textures.size());
        pTextureDesc->setMipmapLevelCount(std::log2(textures[0]->width()));
        
        pTexture = pDevice->newTexture(pTextureDesc.get());
        
        // Use Metal to efficiently blit textures into the texture array:
        
        auto pAutoreleasePool = NS::TransferPtr(NS::AutoreleasePool::alloc()->init());
        
        MTL::CommandBuffer* pCmd       = pCommandBuffer;
        MTL::BlitCommandEncoder* pBlit = pCmd->blitCommandEncoder();
        
        for (size_t i = 0; i < textures.size(); ++i)
        {
            
            auto w = textures[i]->width();
            auto h = textures[i]->height();
            auto d = textures[i]->depth();
            
            pBlit->copyFromTexture(textures[i].get(), 0, 0, MTL::Origin(0,0,0), MTL::Size(w,h,d),
                                   pTexture, i, 0, MTL::Origin(0,0,0));
            
        }
        
        pBlit->generateMipmaps(pTexture);
        
        pBlit->endEncoding();
    }

    return pTexture;
}
